////////////////////////////////////////////////////////
//
//  sample class of bond order potential for muli-body force  //
//
/////////////////////////////////////////////////////////

#include "NBonds.h"
#include "mod.h"
#include "omptune.h"
#include <math.h>
#include <stdio.h>




/*   note:
	AFSIv7_CUTOFF_SHORT̒lɂBulk Modulus͑傫ω
*/



NBonds::NBonds(int maxcount) :
		m_cutoff(1.0),
		m_cutoff2(1.0),
		m_pLCL(NULL){



	// allocate work array for openmp parallel calculation //
	AllocateBuffer(maxcount);
	
}

NBonds::~NBonds(){
	
	
	// delete work array for openmp parallel calculation //
	delete [] m_bvector;
	delete [] m_bvecStartEnd;
	
	delete [] m_num_bonds;

#ifdef _OPENMP

	delete [] sh_aforce;
//	delete [] sh_apot;
	
#endif	//_OPENMP


}

double NBonds::GetCutoffLength(){
	return m_cutoff;
}

double NBonds::GetForceRange(){
	return GetCutoffLength();
}

// allocate work array for openmp parallel calculation //
void NBonds::AllocateBuffer(int maxcount){

	//m_ijmax = maxcount * MAX_NUM_BOND;
	m_bvecStartEnd = new BVEC_REGION[maxcount];
	m_bvector = new BVECTOR[maxcount * MAX_NUM_BOND];
	
	
	m_num_bonds = new int[maxcount];
	
#ifdef _OPENMP
	int num_threads = omp_get_max_threads();
	sh_aforce = new vec3d[maxcount];
//	sh_apot = new double[maxcount * num_threads];
	
#endif	//_OPENMP

}


void NBonds::SetParticlePointers(AtomArray* atoms){
	m_knd = atoms->knd;
	m_r = atoms->r;
	m_p = atoms->p;
	m_f = atoms->f;
	m_mass = atoms->mass;
	m_pot = atoms->pot;
	
	
#ifndef _OPENMP
	
	// if openmp is not enable, calculated force and energy are directly written to user's array, //
	sh_aforce = m_f;
//	sh_apot = m_pot;

	// while if openmp is enable, force and energy are temporarily written to sh_aforce and sh_apot, //
	// which is allocated in AllocateBuffer(), after that  sh_aforce and sh_apot are marged into     //
	// user's array in Force(). //

#endif	//_OPENMP

}

// clear (zero fill) the openmp work arrays. //
inline void NBonds::Reset(int count){
	
#ifdef _OPENMP
	/*
	{//simd
		int sz = (sizeof(vec3d) / sizeof(double)) * count; 
		double* __restrict re_af = (double*)sh_aforce;

		#pragma omp for nowait
		for(int i = 0; i < sz; i++){
			re_af[i] = 0.0;
		}
	}
	*/
	/*
	{//simd

		int* __restrict re_nb = m_num_bonds;

		#pragma omp for
		for(int i = 0; i < count; i++){
			re_nb[i] = 0;
		}
	}
	*/
#else

	
#endif	// _OPENMP
	
	
}


void NBonds::Force(int count, int noloop_count){

// Program is parallelized in this function. //
// Users should not consider openmp parallelizaton in main loop. //
//#pragma omp parallel
{

DEBUG_TM_START;

	// reset work arrays //

	Reset(count);

DEBUG_TM_SET(0);

	// make the linked list //
//	m_pLCL->CalcLinkedList(count, m_r, m_knd);

DEBUG_TM_SET(1);


	// judge and create the bonding list (m_bvector and m_bvecStartEnd) //
	BondListDualDirectS(count - noloop_count);

DEBUG_TM_SET(2);

	// set the array pointer for eash threads ==================== //
#ifdef _OPENMP
	int th = omp_get_thread_num();
#else
	int th = 0;
#endif

	vec3d* a_force = m_f;//sh_aforce + count * th;
	double* a_pot = m_pot;//sh_apot + count * th;
	// ==================== set the array pointer for eash threads //

	mat33d a_Fr_m;
	a_Fr_m.Clear();

	// calculation force and potential, which are written into openmp work array here.//
	ForceByBondList(count - noloop_count, a_force, sh_aforce, a_pot, &a_Fr_m);

DEBUG_TM_SET(3);

	//SingleAtomEnergy(count, a_pot);

DEBUG_TM_SET(4);

	// marge the fore and potentail ========================== //

/*
#ifdef _OPENMP
	// this barrier is necessary before marging force and potential ..
	#pragma omp barrier

//	int t_end = count * omp_get_num_threads();

	{//simd
		vec3d* __restrict re_f = m_f;
		const vec3d* __restrict re_af = sh_aforce;

		#pragma omp for
		for(int i = 0; i < count; i++){
			re_f[i] += re_af[i];
		}
	}

		// marge stress tensor also //
	omptuneReductionSum(&a_Fr_m);
	#pragma omp master
	{
		m_Fr_m = a_Fr_m;
	}

#else	// no-openmp
	m_Fr_m = a_Fr_m;
#endif
*/
	
DEBUG_TM_SET(5);
	
}// finish openmp parallelization //

}




// judge and create the bonding list (m_bvector and m_bvecStartEnd) //
void NBonds::BondListDualDirectS(int count){
	// local variable ------------------------------;

	const double cutoff2 = m_cutoff2;

	LCL_INFO info;
	vec3d periodic_vec[27];
	m_pLCL->GetCellInfo(&info, periodic_vec);
	
	int cell_wx = info.cell_x;
	int cell_wy = info.cell_y;
	int cell_wz = info.cell_z;
	int* icell = info.icell;
	int* first_in_cell = info.first_in_cell;
	int* inext = info.inext;
	
	int cell_wxwy = cell_wx * cell_wy;
	int cell_wxwywz = cell_wxwy * cell_wz;

DEBUG_TM_SET(1);


#ifdef _OPENMP
	// shift array pointer for each thread //
	int th = omp_get_thread_num();
	int thread_count = omp_get_num_threads();
	int bond_index = th * ((count * MAX_NUM_BOND)/thread_count) + 1;
#else
	int bond_index = 1;
#endif


	//int max_num = 0;
	

	//openmp parallelization for atoms//
	#pragma omp for //schedule(static,1)	
	for(int i = 0; i < count; i++){
		vec3d r_i = m_r[i];
		int ic = icell[i];
		int ki = m_knd[i];
		m_bvecStartEnd[i].startpoint = bond_index;	//start point of i-th particle's bond array//


		int icell_z = ic / cell_wxwy;
		int nz[3] = {0, 0, 0};
		int tz[3] = {-cell_wxwy, 0, cell_wxwy};
		
		if(icell_z == 0){
			nz[0] = 18;	//2 * (3 * 3)
			tz[0] += cell_wxwywz;
		}
		if (icell_z == cell_wz - 1){
			nz[2] = 9;	//1 * (3 * 3)
			tz[2] -= cell_wxwywz;
		}

		int icell_y = (ic / cell_wx) % cell_wy;
		int ny[3] = {0, 0, 0};
		int ty[3] = {-cell_wx, 0, cell_wx};
		if(icell_y == 0){
			ny[0] = 6;	//2 * 3
			ty[0] += cell_wxwy;
		}
		if (icell_y == cell_wy - 1){
			ny[2] = 3;	//1 * 3
			ty[2] -= cell_wxwy;
		}

		int icell_x = ic % cell_wx;
		int nx[3] = {0, 0, 0};
		int tx[3] = {-1, 0, 1};
		if(icell_x == 0){
			nx[0] = 2;
			tx[0] = (cell_wx - 1);
		}
		if (icell_x == cell_wx - 1){
			nx[2] = 1;
			tx[2] = (1 - cell_wx);
		}

		// loop for adjoin cell and same cell//
		//int num_t = 0;
		for(int iz = 0; iz < 3; iz++){
			for(int iy = 0; iy < 3; iy++){
				for(int ix= 0; ix < 3; ix++){

					int nxyz = nx[ix] + ny[iy] + nz[iz];
					int target_cell = ic + tx[ix] + ty[iy] + tz[iz];
					
					//if(target_cell == ic) continue;
					
					vec3d rb_i;
					//if(nxyz){//قȂZ//(PperiodicEׂȂɂȂĂ܂Ă)
					//if(target_cell != ic){//قȂZ//(unit cell̍ۂɂׂĂZɂȂĂ܂)
					if(((ix & (iy & iz)) & 0x1) == 0){//(unit cellł)
					
						rb_i = r_i - *(periodic_vec + nxyz);
						for (int j = first_in_cell[target_cell]; j != -1; j = inext[j]){
							vec3d dd = rb_i - m_r[j];

							double r2 = dd * dd;

							// judge the relative distance //
							if (r2 <= cutoff2) {
		
								// add to bonding array //
								BVECTOR& bv_i = m_bvector[bond_index];

								bv_i.pairIndex = j;
								bv_i.rij = dd;
								bv_i.dr = sqrt(r2);
		
								// increment of index //
								bond_index ++;
							}
							//AddBondByCutOffDual(&bond_index, i,j,ki,m_knd[j],dd);			
						//	num_t++;
						}
						
					
					}else{//Z//
						rb_i = r_i;
						for (int j = first_in_cell[target_cell]; j != -1; j = inext[j]){
							if(i == j){
								//x܂ł13/27ZƓZ̑OI//
								m_bvecStartEnd[i].twobodyendpoint = bond_index;	//end point for two-body force of i-th particle's bond//
							}else{
								vec3d dd = rb_i - m_r[j];
								double r2 = dd * dd;

								// judge the relative distance //
								if (r2 <= cutoff2) {
		
									// add to bonding array //
									BVECTOR& bv_i = m_bvector[bond_index];

									bv_i.pairIndex = j;
									bv_i.rij = dd;
									bv_i.dr = sqrt(r2);
		
									// increment of index //
									bond_index ++;
								}
								//AddBondByCutOffDual(&bond_index, i,j,ki,m_knd[j],dd);			
							//	num_t++;
							}
						}
					}
		

				}
			}
		}//iz//
		//if(max_num < num_t){ max_num = num_t;}


		m_bvecStartEnd[i].endpoint = bond_index;	//end point of i-th particle's bond//

	}//end of for i//

	//printf("max_num  = %d\n",max_num);

}

// calculate force and potential //
void NBonds::ForceByBondList(int count, vec3d* __restrict a_force, vec3d* __restrict sh_force, double* __restrict a_pot, mat33d* __restrict a_Fr_m){
	
	const int* __restrict re_knd = m_knd;
	const BVECTOR* re_bvector = m_bvector;
	int* __restrict re_num_bonds = m_num_bonds;
#ifdef _OPENMP
	int th = omp_get_thread_num();
	int thread_count = omp_get_num_threads();
#else
	int th = 0;
	int thread_count = 1;
#endif
	int i_start = (th * count)/thread_count;
	int i_end = ((th + 1) * count)/thread_count;
	
	for (int i = i_start; i < i_end; i++){

	//loop of atom is parallelized by openmp //
	//#pragma omp for //schedule(static,1)	
	//for (int i = 0; i < count; i++){
		
		int ki = re_knd[i];
		int num_bonds = 0;

		// loop of j-th atom which forms i-j bond //
		const int twobody_endpoint = m_bvecStartEnd[i].twobodyendpoint;
		const int bv_idx_i_start = m_bvecStartEnd[i].startpoint;
		const int bv_idx_i_end = m_bvecStartEnd[i].endpoint;

		for(int bv_idx_i = bv_idx_i_start; bv_idx_i < bv_idx_i_end; bv_idx_i ++){
			const BVECTOR& bv_i = re_bvector[bv_idx_i];
			int j = bv_i.pairIndex;
			//if(j < i) continue;		// prevent double count of bond //
			
			int kj = re_knd[j];
			
			if(ki == KIND_W){
				if(kj == KIND_W){
					num_bonds ++;
				}
			}else{
				num_bonds ++;
				
			}
		}

		m_num_bonds[i] = num_bonds;
	}
	
}



					

inline void NBonds::AddFr(vec3d force_ij, vec3d r_ij, mat33d* a_Fr_m){
	a_Fr_m->a += force_ij * r_ij.x;
	a_Fr_m->b += force_ij * r_ij.y;
	a_Fr_m->c += force_ij * r_ij.z;
}

void NBonds::GetFr(mat33d* a_Fr_m){
		*a_Fr_m = m_Fr_m;

}

void NBonds::SetParams(const double* params){
	m_cutoff = *params;
	m_cutoff2 = m_cutoff * m_cutoff;
}

void NBonds::GetInfo(int mode, void* info){

	switch(mode){
	case 3:
		*(int**)info = m_num_bonds;
		break;
	case 1:
		*(BVECTOR**)info = m_bvector;
		break;
	case 2:
		*(BVEC_REGION**)info = m_bvecStartEnd;
		break;
	}

}

